DrDamage = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceHook-2.1")
local L = AceLibrary("AceLocale-2.2"):new("DrDamage")
local BS = AceLibrary("Babble-Spell-2.2")
local SEA = AceLibrary("SpecialEvents-Aura-2.0")
local GT = AceLibrary("Gratuity-2.0")
local IBL

local defaults = {
	--Calculation
	HitCalc = false,
	HitTarget = true,
	TargetLevel = -1,
	TargetPlayer = false,
	SpellDamage = 0,
	Healing = 0,
	ShadowDamage = 0,
	FireDamage = 0,
	NatureDamage = 0,
	FrostDamage = 0,
	ArcaneDamage = 0,
	HolyDamage = 0,
	ItemCrit = 0,
	ItemHit = 0,
	ManualDmg = false,
	ManualPlus = false,	
	--Actionbar:
	DisplayType = "Avg",
	TrueAvg = true,
	FontSize = 8,
	FontXPosition = 0,
	FontYPosition = 0,
	CastsLeft = false,
	--Tooltip:
	Display = true,
	PlusDmg = true,
	DispCrit = true,
	DispHit = true,
	AvgHit = true,
	AvgCrit = true,
	Ticks = true,
	Total = true,
	Extra = true,
	Next = true,
	DPS = true,
	DPM = true,
	Doom = true,
}

DrDamage:RegisterDB("DrDamageDB");
DrDamage:RegisterDefaults("profile", defaults )

function DrDamage:OnInitialize()	
end

local changesDMG = {
    "CHARACTER_POINTS_CHANGED",
    "PLAYER_ALIVE",
}

function DrDamage:CHARACTER_POINTS_CHANGED()
    	DrDamage:DmgUpdateTalents()
    	DrDamage:UpdateAB()
end

function DrDamage:PLAYER_ALIVE()
    	DrDamage:DmgUpdateTalents()
    	DrDamage:UpdateAB()
end

local changesAB = {
    "ACTIONBAR_PAGE_CHANGED",
    "ACTIONBAR_HIDEGRID",
    "UNIT_INVENTORY_CHANGED",
    "PLAYER_TARGET_CHANGED",
    "UNIT_AURA",
}

function DrDamage:UNIT_AURA( unit )

	if unit == "target" or unit =="player" then
		
		if not self:IsEventScheduled("UpdatingAB") then
	    		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.5, self)
	    	end
	end
end

function DrDamage:ACTIONBAR_PAGE_CHANGED()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self)
	end
end

function DrDamage:ACTIONBAR_HIDEGRID()
	
	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self)
	end
end

function DrDamage:PLAYER_TARGET_CHANGED()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self)
	end
end

function DrDamage:UNIT_INVENTORY_CHANGED( unit )
	
	if unit == "player" then
		if not self:IsEventScheduled("UpdatingAB") then
			self:ScheduleEvent("UpdatingAB", self.UpdateAB, 2, self)
		end
	end
end

local talents = {}
local spellInfo = {}
local talentEffects = {}
local _G

function DrDamage:OnEnable()

	if not DrDamage:DetermineClass() then
		--self:Print( "Your class is not supported by Dr. Damage." )
		do return end
	end
	
	talents = {}
	_G = getfenv(0)
	
	DrDamage:Hook( GameTooltip, "SetAction", true)
	DrDamage:SecureHook( GameTooltip, "SetSpell" )
	DrDamage:DmgUpdateTalents()
	
    	for i = 1, #changesDMG do
    		
    		if changesDMG[i] then
    			self:RegisterEvent( changesDMG[i] )
    		end
  	end
	
	if DrDamage.db.profile.DisplayType ~= "Off" then
    		
    		DrDamage:UpdateAB()
    	
    		for i = 1, #changesAB do
    	
    			if changesAB[i] then
    				self:RegisterEvent( changesAB[i] )
    			end
    		end
  	end  	
end

function DrDamage:OnDisable()

	DrDamage:UpdateAB( true )
	talents = nil
	spellInfo = nil
	talentEffects = nil
	_G = nil
	
end

function dGet (n)
	return function () return DrDamage.db.profile[n] 
	end
end
function dSet (n, v)
	return function (v) DrDamage.db.profile[n] = v; DrDamage:UpdateAB(); 
	end
end

local DisplayTypes = { "Avg", "DPS", "DPSC", "Off" }



--LOCALE:
local options = { 
	type='group',
	args = {
		ActionBar = {
			type = "group", desc = "Options for the actionbar", name="ActionBar",
			args = {
				DisplayType = {
					name = "Display",
					type = "text",
					desc = "Choose what to display on the actionbar",
					validate = DisplayTypes,
					get = dGet("DisplayType"),
					set = dSet("DisplayType", value),
					order = 70,
				},
				TrueAvg = {
					type = 'toggle',
					name = "Include crits/DoT effects.",
					desc = "Toggles if additional effects are included in the average",
					order = 71,
					get = dGet( "TrueAvg" ),
					set = dSet( "TrueAvg", value ),				
				},
        			Update = {
            				name = "Update",
            				desc = "Forces update to actionbar",
            				type = "execute",
            				order = 75,
            				func = "UpdateAB",
    				}, 
				FontSize = {
					type = 'range',
					name = "Actionbar font size",
					min = 6,
					max = 14,
					step = 1,
					desc = "Set the font size",
					order = 73,
					get = dGet("FontSize"),
					set = dSet("FontSize", value),
				},
				FontXPosition = {
					type = 'range',
					name = "Actionbar font horizontal position",
					min = -5,
					max = 5,
					step = 1,
					desc = "Set the font horizontal position",
					order = 74,
					get = dGet("FontXPosition"),
					set = dSet("FontXPosition", value),
				},
				FontYPosition = {
					type = 'range',
					name = "Actionbar font vertical position",
					min = -15,
					max = 30,
					step = 1,
					desc = "Set the font vertical position",
					order = 75,
					get = dGet("FontYPosition"),
					set = dSet("FontYPosition", value),
				}, 				
				CastsLeft = {
					type = 'toggle',
					name = "Healing casts left",
					desc = "Toggles amount of casts left to display on actionbar",
					order = 80,
					get = dGet( "CastsLeft" ),
					set = dSet( "CastsLeft", value ),
				},				
			},
		},
		Tooltip = {
			type = "group", desc = "Options for the tooltips", name="Tooltip",
			args = {
				Display = {
					type = 'toggle',
					name = "Tooltip on/off",
					desc = "Toggles the tooltip info on/off",
					order = 50,
					get = dGet( "Display" ),
					set = dSet( "Display", value ),
				},			
				PlusDmg = {
					type = 'toggle',
					name = "Show +dmg/+healing",
					desc = "Toggles displaying of +dmg/+healing",
					order = 52,
					get = dGet( "PlusDmg" ),
					set = dSet( "PlusDmg", value ),
				},
				DispCrit = {
					type = 'toggle',
					name = "Show crit %",
					desc = "Toggles displaying of crit %",
					order = 54,
					get = dGet( "DispCrit" ),
					set = dSet( "DispCrit", value ),
				},
				DispHit = {
					type = 'toggle',
					name = "Show hit %",
					desc = "Toggles displaying of hit %.",
					order = 56,
					get = dGet( "DispHit" ),
					set = dSet( "DispHit", value ),
				},				
				AvgHit = {
					type = 'toggle',
					name = "Show avg + hit range",
					desc = "Toggles displaying of avg hit",
					order = 58,
					get = dGet( "AvgHit" ),
					set = dSet( "AvgHit", value ),
				},
				AvgCrit = {
					type = 'toggle',
					name = "Show avg + crit range",
					desc = "Toggles displaying of avg crit",
					order = 60,
					get = dGet( "AvgCrit" ),
					set = dSet( "AvgCrit", value ),
				},
				Ticks = {
					type = 'toggle',
					name = "Show per tick/hit",
					desc = "Toggles displaying of per hit/tick values",
					order = 62,
					get = dGet( "Ticks" ),
					set = dSet( "Ticks", value ),
				},
				Total = {
					type = 'toggle',
					name = "Show avg total dmg/heal",
					desc = "Toggles displaying of average in total values",
					order = 64,
					get = dGet( "Total" ),
					set = dSet( "Total", value ),
				},				
				Extra = {
					type = 'toggle',
					name = "Show extra dmg/heal (DoT, Ignite, Chained)",
					desc = "Toggles displaying of extra info",
					order = 66,
					get = dGet( "Extra" ),
					set = dSet( "Extra", value ),
				},
				Next = {
					type = 'toggle',
					name = "Show next values (+1% crit, +10 dmg)",
					desc = "Toggles displaying of +1% crit, +10 dmg values",
					order = 68,
					get = dGet( "Next" ),
					set = dSet( "Next", value ),
				},
				DPS = {
					type = 'toggle',
					name = "Show DPS/HPS",
					desc = "Toggles displaying of DPS/HPS",
					order = 70,
					get = dGet( "DPS" ),
					set = dSet( "DPS", value ),
				},
				DPM = {
					type = 'toggle',
					name = "Show DPM/HPM",
					desc = "Toggles displaying of DPM/HPM",
					order = 72,
					get = dGet( "DPM" ),
					set = dSet( "DPM", value ),
				},
				Doom = {
					type = 'toggle',
					name = "Show Damage/healing until OOM",
					desc = "Toggles displaying of damage/healing until OOM",
					order = 74,
					get = dGet( "Doom" ),
					set = dSet( "Doom", value ),
				},				
			},
		},		
		Calculation = {
			type = "group", desc = "Calculation options", name="Calculation",
			args = {
				HitCalc = {			
					type = 'toggle',
					name = "+Hit calculation",
					desc = "Toggles +hit calculation effects into averages on/off",
					order = 65,
					get = dGet( "HitCalc" ),
					set = dSet( "HitCalc", value ),
				},
				HitTarget = {			
					type = 'toggle',
					name = "Hit calculation by target level",
					desc = "Toggles +hit calculation by target level.",
					order = 66,
					get = dGet( "HitTarget" ),
					set = dSet( "HitTarget", value ),
				},
				TargetLevel = {
					type = 'range',
					name = "Manual target level setting",
					desc = "Manual set of target level compared to your level. Turn off hit by target level to use this.",
					min = -10,
					max = 10,
					step = 1,
					order = 67,
					get = dGet( "TargetLevel" ),
					set = dSet( "TargetLevel", value ),
				},
				TargetPlayer = {
					type = 'toggle',
					name = "Manual target is a player.",
					desc = "Toggles if your manually set target is a player.",
					order = 68,
					get = dGet( "TargetPlayer" ),
					set = dSet( "TargetPlayer", value ),
				},				
				ManualDmg = {
					type = 'toggle',
					name = "Manual variables",
					desc = "Allows you to manually set damage/healing properties",
					order = 71,
					get = dGet( "ManualDmg" ),
					set = dSet( "ManualDmg", value ),
				},
				ManualPlus = {
					type = 'toggle',
					name = "Manual damage adds to detected",
					desc = "Allows you to set if manual values should add to detected values",
					order = 72,
					get = dGet( "ManualPlus" ),
					set = dSet( "ManualPlus", value ),
				}, 
				Healing = {
					type = 'range',
					name = "Healing",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your +healing",
					order = 79,
					get = dGet("Healing"),
					set = dSet("Healing", value),
				},				
				SpellDamage = {
					type = 'range',
					name = "Spell Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your spell damage",
					order = 80,
					get = dGet("SpellDamage"),
					set = dSet("SpellDamage", value),
				},		
				FireDamage = {
					type = 'range',
					name = "Fire Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your fire +damage",
					order = 81,
					get = dGet("FireDamage"),
					set = dSet("FireDamage", value),		
		
				},
				ShadowDamage = {
					type = 'range',
					name = "Shadow Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your shadow +damage",
					order = 82,
					get = dGet("ShadowDamage"),
					set = dSet("ShadowDamage", value),		
				},
				FrostDamage = {
					type = 'range',
					name = "Frost Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your frost +damage",
					order = 83,
					get = dGet("FrostDamage"),
					set = dSet("FrostDamage", value),		
				},
				NatureDamage = {
					type = 'range',
					name = "Nature Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your nature +damage",
					order = 84,
					get = dGet("NatureDamage"),
					set = dSet("NatureDamage", value),		
				},
				ArcaneDamage = {
					type = 'range',
					name = "Arcane Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your arcane +damage",
					order = 85,
					get = dGet("ArcaneDamage"),
					set = dSet("ArcaneDamage", value),		
				},
				HolyDamage = {
					type = 'range',
					name = "Holy Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your holy +damage",
					order = 85,
					get = dGet("HolyDamage"),
					set = dSet("HolyDamage", value),		
				},					
				ItemCrit = {
					type = 'range',
					name = "Item Crit Bonus",
					min = 0,
					max = 1,
					isPercent = true,
					desc = "Input your crit bonus from items",
					order = 88,
					get = dGet("ItemCrit"),
					set = dSet("ItemCrit", value),					
				},
				ItemHit = {
					type = 'range',
					name = "Item Hit Bonus",
					min = 0,
					max = 1,
					isPercent = true,
					desc = "Input your hit bonus from items",
					order = 89,
					get = dGet("ItemHit"),
					set = dSet("ItemHit", value),	
				},
			},
		},
	},
};

DrDamage:RegisterChatCommand( { "/drdmg", "/drdamage" }, options);

function DrDamage:DmgUpdateTalents()

	for k in pairs( talents ) do 
		talents[k] = nil 
	end 
	
	local numTabs = GetNumTalentTabs();
	
	for t=1, numTabs do
	
    		local numTalents = GetNumTalents(t);
    		  		
	    	for i=1, numTalents do
	    	
	    		local nameTalent, _, _, _, currRank, maxRank= GetTalentInfo(t,i);
	    		
	    		if BS:HasReverseTranslation( nameTalent ) then
	    			nameTalent = BS:GetReverseTranslation( nameTalent )
	    		end
	    		
	    	    	if currRank ~= 0 then
	    	    			    	    
	    	    		if talentEffects[nameTalent] then
	    	    			table.insert( talents, { Name = nameTalent, Rank = currRank } )
	    	    			--self:Print("Talent name: " .. nameTalent .. " Points: " .. currRank .. "/" .. maxRank )
	    	    		end
	    	    	end	    	    
	    	end
	end
end


function DrDamage:DetermineClass()

	_, englishClass = UnitClass("player");
	local nomatch
	local deleted

	if englishClass == "WARLOCK" then
		if type( DrDamage.WarlockData ) == "function" then
			spellInfo, talentEffects = DrDamage:WarlockData()
		else
			deleted = true
		end
	elseif englishClass == "MAGE" then
		if type( DrDamage.MageData ) == "function" then
			spellInfo, talentEffects = DrDamage:MageData()
		else
			deleted = true
		end		
	elseif englishClass == "PRIEST" then
		if type( DrDamage.PriestData ) == "function" then
			spellInfo, talentEffects = DrDamage:PriestData()
		else
			deleted = true
		end		
	elseif englishClass == "SHAMAN" then
		if type( DrDamage.ShamanData ) == "function" then
			spellInfo, talentEffects = DrDamage:ShamanData()
		else
			deleted = true
		end	
	elseif englishClass == "DRUID" then
		if type( DrDamage.DruidData ) == "function" then
			spellInfo, talentEffects = DrDamage:DruidData()
		else
			deleted = true
		end	
	else
		nomatch = true
	end
	
	if deleted then
		self:Print( "The data file for " .. englishClass:lower() .. "s has been deleted." )
		return false
	elseif nomatch then
		return false
	else
		return true	
	end
end

function findpattern(text, pattern, start)
	if text and pattern and string.find( text, pattern, start) then
		return string.sub(text, string.find(text, pattern, start))
	else
		return " "
	end
end

function DrDamage:SetSpell( frame, slot )

	if not DrDamage.db.profile.Display then
		do return end
	end
		
	GT:SetSpell( slot, BOOKTYPE_SPELL )
	local spellName, spellRank = GetSpellName( slot, BOOKTYPE_SPELL );
	
	if spellRank and string.len( spellRank ) >= 6 then
		spellRank = tonumber( findpattern( spellRank, "%d+" ))
	end
	
	if not spellRank then
		spellRank = "None"
	end
	
	DrDamage:DoTooltip( frame, spellName, spellRank, true )

end

function DrDamage:SetAction( frame, slot )

	self.hooks[frame].SetAction(frame, slot)
	
	if not DrDamage.db.profile.Display then
		do return end
	end
	
	local gtype, _, _ = GetActionInfo( slot )
	
	if gtype == "spell" then
	
		local spellName = getglobal("GameTooltipTextLeft1"):GetText()
		local spellRank = getglobal("GameTooltipTextRight1"):GetText()
	
		if spellRank and string.len( spellRank ) >= 6 then
			spellRank = tonumber( findpattern( spellRank, "%d+" ))
		else
			spellRank = "None"
		end
	
		GT:SetAction( slot )
		DrDamage:DoTooltip( frame, spellName, spellRank, true )
		
	elseif gtype == "macro" then
	
		local macroName = getglobal("GameTooltipTextLeft1"):GetText()
		local spellName
		local spell
				
		if macroName and string.find( macroName, '%(%d+%)') then
			spellName = string.sub( macroName, 1, string.find( macroName, "%(" ) - 1 )
		else
			spellName = macroName
		end
				
		if spellName then
			
			spell = spellInfo[spellName]
			
			if not spell then
					
				for index, cSpell in pairs( spellInfo ) do
						
  					if (type( index ) == "string" ) then
  							
  						if string.find( index, spellName ) then
  								
  							local position = string.find( index, spellName )
  								
  							if position == 1 then
  								spell = cSpell
  								spellName = index
  							end
  						end
					end	
				end
			end
		end
								
		if spell then
										
			local macroRank = findpattern( macroName, '%(%d+%)')
						
			if macroRank then
				macroRank = tonumber( string.sub( macroRank, 2, string.len( macroRank ) - 1 ))
			end
						
			local tableSpell
						
			if macroRank then
				if macroRank > 0 then
					tableSpell = spell[macroRank]
				end
				
				if tableSpell then
					DrDamage:DoTooltip( frame, spellName, macroRank, false )
				end	
			else
				local spellRank = 0
				local playerLevel = UnitLevel("player")
						
				for index, ispell in pairs( spell ) do
						
  					if (type( index ) == "number") and index > spellRank then
  							
  						if ispell.spellLevel <= playerLevel then
							spellRank = index
						end
     					end
				end
							
				if spellRank > 0 then
					tableSpell = spell[spellRank]
				end
				
				if tableSpell then
					DrDamage:DoTooltip( frame, spellName, spellRank, false )
				end					
			end	
		end	
	end
end


local TooltipData = {}

function DrDamage:DoTooltip( frame, spellName, spellRank, readTooltip )

	if spellName and spellRank then
  	
		if spellInfo[spellName] then
		
			local spell = spellInfo[spellName]
			local baseSpell = spell[0]
			local tableSpell = spell[spellRank]
			
			if not tableSpell then
				tableSpell = spell["None"]
			end
			
			if tableSpell and not baseSpell.NoTooltip then
				
				local healingSpell, AvgTotal, DPS, DPSC, AvgDmg, TooltipData = DrDamage:RawNumbers ( tableSpell, spellName )
				local DPM
				local DOOM
				local ManaCost
	
				if readTooltip then
					_, _, ManaCost = GT:Find(L["(%d+) Mana"], 2, 2)
				end				
				
				if ManaCost then
				
					ManaCost = tonumber( ManaCost )
					
					if ManaCost > 0 then
						
						DPM = DrDamage:Round( AvgTotal / ManaCost, 1 )
					
						local PlayerMana = UnitManaMax("player")
						if PlayerMana then
							DOOM = DrDamage:Round( DPM * PlayerMana )
						end
					end					
					
				end
				
				--LOCALE:
				frame:AddLine(" ")
				
				if DrDamage.db.profile.PlusDmg then
					if healingSpell then
						frame:AddDoubleLine("+Healing: ", TooltipData.SpellDmg .. " (x" .. TooltipData.SpellDmgM .. ")", 1, 1, 1, 0, 0.7, 0  )
					else
						frame:AddDoubleLine("+Dmg: ", TooltipData.SpellDmg .. " (x" .. TooltipData.SpellDmgM .. ")", 1, 1, 1, 0, 0.7, 0  )
					end
				end
  				
  				if DrDamage.db.profile.DispCrit and TooltipData.CritRate > 0 then
  					frame:AddDoubleLine("Crit:", TooltipData.CritRate .. "%", 1, 1, 1, 0, 0.7, 0 )
  				end
  				
  				if DrDamage.db.profile.DispHit and not baseSpell.Unresistable and not healingSpell then
  					frame:AddDoubleLine("Hit:", TooltipData.HitRate .. "%", 1, 1, 1, 0, 0.7, 0 )
  				end  				
  				
  				if DrDamage.db.profile.AvgHit then
  					frame:AddDoubleLine("Avg (x".. TooltipData.DmgM .. "):", AvgDmg .. " (".. TooltipData.MinDmg .."-".. TooltipData.MaxDmg ..")", 1, 1, 1, 0.8, 0.8, 0.9 )
  				end
  				
  				if DrDamage.db.profile.AvgCrit and TooltipData.AvgCrit > 0 and not baseSpell.sHits then
  					frame:AddDoubleLine("Avg Crit:", TooltipData.AvgCrit .. " (".. TooltipData.MinCrit .."-".. TooltipData.MaxCrit ..")", 1, 1, 1, 0.8, 0.8, 0.9 )
  				end
  				
  				if DrDamage.db.profile.Ticks then 
  					if TooltipData.PerHit and baseSpell.sHits then
  						frame:AddDoubleLine( "Hits:", TooltipData.PerHit[2] .. "x ~" .. TooltipData.PerHit[1], 1, 1, 1, 0.8, 0.8, 0.9 )
  					elseif TooltipData.PerTick and baseSpell.sTicks then
  						frame:AddDoubleLine( "Ticks:", TooltipData.PerTick[2] .. "x ~" .. TooltipData.PerTick[1], 1, 1, 1, 0.8, 0.8, 0.9 )
  					end
  				end
  				
  				if DrDamage.db.profile.Extra then 
  				
  					if TooltipData.ChainDmg > 0 then
  						frame:AddDoubleLine("Avg Chain Effect:", TooltipData.ChainDmg, 1, 1, 1, 0.8, 0.8, 0.9  )
  					end
  				
  					if TooltipData.IgniteDmg > 0 then
  						frame:AddDoubleLine("Ignite:", TooltipData.IgniteDmg, 1, 1, 1, 0.8, 0.8, 0.9 )
  					end
  				
   					if TooltipData.DotDmg > 0 then
  						frame:AddDoubleLine("DoT:", TooltipData.DotDmg, 1, 1, 1, 0.8, 0.8, 0.9 )
  					end
  				end  				
  				
  				if DrDamage.db.profile.Total and AvgTotal > AvgDmg then
  					frame:AddDoubleLine("Avg Total:", AvgTotal, 1, 1, 1, 0.8, 0.8, 0.9  )
  				end
  				  				
  				if DrDamage.db.profile.Next then

  					if TooltipData.NextCrit > 0 then
  						if healingSpell then
  							frame:AddDoubleLine("+1% Crit:", "+" .. TooltipData.NextCrit .. " (= " .. DrDamage:Round( ( TooltipData.NextCrit * 10 ) / TooltipData.NextTenDmg, 1 ) .. " +heal)", 1, 1, 1, 0.3, 0.6, 0.5 )	
  						else
  							frame:AddDoubleLine("+1% Crit:", "+" .. TooltipData.NextCrit .. " (= " .. DrDamage:Round( ( TooltipData.NextCrit * 10 ) / TooltipData.NextTenDmg, 1 ) .. " +dmg)", 1, 1, 1, 0.3, 0.6, 0.5 )	
  						end
  					end
  					
  					if not healingSpell and not baseSpell.Unresistable then
  						frame:AddDoubleLine( "+1% Hit:", "+" .. TooltipData.NextHit, 1, 1, 1, 0.3, 0.6, 0.5 )
  					end
  				
  					if healingSpell then
  						frame:AddDoubleLine( "+10 Healing:", "+" .. TooltipData.NextTenDmg, 1, 1, 1, 0.3, 0.6, 0.5 )
  					else
  						frame:AddDoubleLine( "+10 Dmg:", "+" .. TooltipData.NextTenDmg, 1, 1, 1, 0.3, 0.6, 0.5 )
  					end
  					
  				end
  				
  				if DrDamage.db.profile.DPS and not baseSpell.NoDPS then
  				
  					if DPSC ~= DPS and not baseSpell.NoDPS then
  						if healingSpell then
  							frame:AddDoubleLine("HPS/HPSC:", DPS .. "/" .. DPSC, 1, 1, 1, 0.8, 0.1, 0.1)
  						else
  							frame:AddDoubleLine("DPS/DPSC:", DPS .. "/" .. DPSC, 1, 1, 1, 0.8, 0.1, 0.1)
      						end
      					else
      				
      						if healingSpell then
      							frame:AddDoubleLine("HPS:", DPS, 1, 1, 1, 0.8, 0.1, 0.1)
      						
      							if TooltipData.DPSCD > 0 then
      								frame:AddDoubleLine("HPS (CD):", TooltipData.DPSCD, 1, 1, 1, 0.8, 0.1, 0.1)
      							end
      						else
      							frame:AddDoubleLine("DPS:", DPS, 1, 1, 1, 0.8, 0.1, 0.1)
      						
      							if TooltipData.DPSCD > 0 then
      								frame:AddDoubleLine("DPS (CD):", TooltipData.DPSCD, 1, 1, 1, 0.8, 0.1, 0.1)
      							end      						
      						end
      					end
      				end
      				
      				if DrDamage.db.profile.DPM and DPM and not baseSpell.NoDPM then
      					if healingSpell then
      						frame:AddDoubleLine("HPM:", DPM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					else
      						frame:AddDoubleLine("DPM:", DPM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					end
      				end
      				
      				if DrDamage.db.profile.Doom and DOOM and not baseSpell.NoDoom then
      					if healingSpell then
      						frame:AddDoubleLine("Healing until OOM:", DOOM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					else
      						frame:AddDoubleLine("Damage until OOM:", DOOM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					end
      				end
      				
				for k in pairs( TooltipData ) do 
					TooltipData[k] = nil 
				end      				
      				
      				frame:Show()			
			end  			
		end	  
	end
end
	
local ABtable
	
function DrDamage:UpdateAB( disable )

	local BT3 = false
	local Bongos = false
	local DAB = false
	local NAB = false
	local Trinity = false
	local CTBar = false
	local blizzAB = false
	
	if IsAddOnLoaded("Bartender3") then
		BT3 = true
		if not ABtable then
			ABtable = Bartender3.actionbuttons
		end
	elseif IsAddOnLoaded( "Bongos" ) then
		Bongos = true
	elseif IsAddOnLoaded("DiscordActionBars") then
		DAB = true
	elseif IsAddOnLoaded("Nurfed_ActionBars") then
		NAB = true
	elseif IsAddOnLoaded("TrinityBars") then
		Trinity = true
	elseif IsAddOnLoaded("CT_BarMod") then
		CTBar = true
		if not ABtable then
			ABtable = CT_BarMod.actionButtonList
		end
	else
		blizzAB = true
		if not ABtable then
			ABtable = { "ActionButton", "BonusActionButton", "MultiBarBottomLeftButton", "MultiBarBottomRightButton","MultiBarRightButton", "MultiBarLeftButton", }
		end
	end
	
	function ProcessButton( button, index, main )
		
		local ABTextChildren = { button:GetChildren() }
		
   		for _,frame in ipairs( ABTextChildren ) do
   				
   			if frame and frame.text and frame:GetID() == 200 then
   				frame.text:Hide()
   			end
   		end
   			
   		ABTextChildren = nil
			
		if button:IsVisible() and not disable then
		
			local id
		
			if BT3 and ABtable[index] and ABtable[index].state == "used" then
				id = ABtable[index].object:PagedID()
			elseif Bongos then
				id = button:GetPagedID()
			elseif DAB then
				id = button:GetActionID()
			elseif NAB or Trinity then
				id = button:GetID()
			elseif CTBar then
				if main then
					id = ActionButton_GetPagedID( button )
				elseif ABtable[index] and ABtable[index].hasAction then
					id = index
				end
			elseif blizzAB then
				id = ActionButton_GetPagedID( button )
			end
			
			if id then
				DrDamage:CheckAction( id, button )
			end
		end
	end
	
	
	if CTBar then
	
		local button
		
		for i = 1, 12 do
			button = _G["ActionButton"..i]
			
			if button then
				ProcessButton( button, i, true )
			end
		end		
		
		for key, value in pairs( ABtable ) do 
			
			button = value.button
			
			if button then
				ProcessButton( button, key )
			end
		end
		

	else
		local button
		
		for i = 1, 120 do
	
			if BT3 then
				button = _G["BT3Button"..i]
			elseif Bongos then
				button = _G["BActionButton"..i]
			elseif DAB then
				button = _G["DAB_ActionButton_"..i]
			elseif NAB then
				button = _G["Nurfed_ActionButton"..i]
			elseif Trinity then
				button = _G["TrinityActionButton"..i]
			elseif blizzAB then
				local abType = 1 + math.floor( ( i - 1 ) / 12 )
				local buttonNr = i - math.floor( ( i - 1 ) / 12 ) * 12
			
				if abType > #ABtable then
					do return end
				end
			
				button = _G[ABtable[abType]..buttonNr]
			end
					
			if button then
				ProcessButton( button, i )
			end
		end
	end
end

function DrDamage:CheckAction( id, button )

	local gtype, pid, _ = GetActionInfo(id)
			
	if gtype == "spell" then
		
		local spellName, spellRank = GetSpellName( pid, BOOKTYPE_SPELL );
			
		if spellInfo[spellName] then
				
			local spell = spellInfo[spellName]
			local tableSpell = spell[ tonumber( findpattern( spellRank, "%d+" ) ) ]
						
			if not tableSpell then
				tableSpell = spell["None"]
			end
		
			if tableSpell then

				local healingSpell, AvgTotal, DPS, DPSC, Avg = DrDamage:RawNumbers ( tableSpell, spellName )
							
				local didCasts = false
				local baseSpell = spell[0]
							
				if DrDamage.db.profile.CastsLeft and healingSpell then
							
					GT:SetAction( id )
					local _, _, ManaCost = GT:Find(L["(%d+) Mana"], 2, 2)
					local PlayerMana = UnitManaMax("player")
							
					if ManaCost then
						ManaCost = tonumber( ManaCost )
						local castsLeft = DrDamage:Round( PlayerMana / ManaCost )
						DrDamage:CreateABtext( button, castsLeft, healingSpell )
						didCasts = true
					end
								
				end
							
				if not didCasts then
							
					local displayType = DrDamage.db.profile.DisplayType
							
					if displayType == "Avg" or baseSpell.NoDPS and displayType ~= "Off" then
						if not DrDamage.db.profile.TrueAvg then
							DrDamage:CreateABtext( button, Avg, healingSpell )
						else
							DrDamage:CreateABtext( button, AvgTotal, healingSpell )
						end
					elseif displayType == "DPS" then
						DrDamage:CreateABtext( button, DPS, healingSpell )
					elseif displayType == "DPSC" then
						DrDamage:CreateABtext( button, DPSC, healingSpell )
					end
				end
			end				
		end
					
	elseif gtype == "macro" then
				
		local macroName = GetActionText( id )
		local spellName
		local spell
				
		if macroName and string.find( macroName, '%(%d+%)') then
			spellName = string.sub( macroName, 1, string.find( macroName, "%(" ) - 1 )
		else
			spellName = macroName
		end
					
		if spellName then
					
			spell = spellInfo[spellName]
					
			if not spell then
					
				for index, cSpell in pairs( spellInfo ) do
						
  					if (type( index ) == "string" ) then
  							
  						if string.find( index, spellName ) then
  								
  							local position = string.find( index, spellName )
  									
  							if position == 1 then
  								spell = cSpell
  								spellName = index
  							end
  						end
					end	
				end
			end
		end
								
		if spell then
					
			local macroRank = findpattern( macroName, '%(%d+%)')
						
			if macroRank then
				macroRank = tonumber( string.sub( macroRank, 2, string.len( macroRank ) - 1 ))
			end
						
			local tableSpell
						
			if macroRank then
				if macroRank > 0 then
					tableSpell = spell[macroRank]
				end
			else
				local spellRank = 0
				local playerLevel = UnitLevel("player")
						
				for index, ispell in pairs( spell ) do
						
  					if (type( index ) == "number") and index > spellRank then
  							
  						if ispell.spellLevel <= playerLevel then
							spellRank = index
						end
     					end
				end
							
				if spellRank > 0 then
					tableSpell = spell[spellRank]
				end
			end
						
			if tableSpell then
						
				local macroText = getglobal(button:GetName() .. "Name")
					
				if macroText then
					macroText:Hide()
				end
				
				if IsAddOnLoaded("CT_BarMod") then
					if button.name then
						button.name:SetText( "" )
					end
				end
				
				local baseSpell = spell[0]
				local healingSpell, AvgTotal, DPS, DPSC, Avg = DrDamage:RawNumbers ( tableSpell, spellName )
				local displayType = DrDamage.db.profile.DisplayType
									
				if displayType == "Avg" or baseSpell.NoDPS and displayType ~= "Off" then
						
						if not DrDamage.db.profile.TrueAvg then
								DrDamage:CreateABtext( button, Avg, healingSpell )
						else
							DrDamage:CreateABtext( button, AvgTotal, healingSpell )
						end								
				elseif displayType == "DPS" then
						DrDamage:CreateABtext( button, DPS, healingSpell )
				elseif displayType == "DPSC" then
						DrDamage:CreateABtext( button, DPSC, healingSpell )
				end
			end				
		end							
	end
end


function DrDamage:CreateABtext( button, cDmg, HealingSpell )

	local ABTextChildren = { button:GetChildren() }
	local found = false
	local positionMod = 0
	cDmg = DrDamage:Round( cDmg )
	
	if cDmg and string.len( cDmg ) then
		positionMod = math.floor( (4 - string.len( cDmg ) ) * 3.5 )
	end
	
   	for _,frame in ipairs( ABTextChildren ) do
   			
   		if frame and frame.text and frame:GetID() == 200 then
   			
   			frame.text:SetPoint("BOTTOMLEFT", frame, "BOTTOM", (DrDamage.db.profile.FontXPosition - 15) + positionMod, (DrDamage.db.profile.FontYPosition + 5) )
   			frame.text:SetFont( STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE" )
   			frame.text:SetText( cDmg )
   			frame.text:Show()
   			found = true
   		end	
   	end
   	
   	ABTextChildren = nil
   		
   	if not found then
   		
		local dpsText = CreateFrame("Frame", "DrDamageText", button )
		dpsText:SetAllPoints( button )
		dpsText:SetFrameLevel(dpsText:GetFrameLevel() + 1)
		dpsText:SetID( 200 )
		dpsText.text = dpsText:CreateFontString( nil, "OVERLAY" )
		dpsText.text:SetFontObject(GameFontNormalSmall)
		dpsText.text:SetPoint("BOTTOMLEFT", dpsText, "BOTTOM", (DrDamage.db.profile.FontXPosition - 15) + positionMod, (DrDamage.db.profile.FontYPosition + 5) )
		dpsText.text:SetFont(STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE")
		
		if HealingSpell then
			dpsText.text:SetTextColor(0.4, 1.0, 0.3)
		else
			dpsText.text:SetTextColor(1, 1, 0.2)
		end
		
		dpsText.text:SetText( cDmg )
		dpsText.text:Show()
	end

end

function DrDamage:SpellCrit()

	local crit = GetSpellCritChanceFromIntellect( "player" ) + GetCombatRatingBonus(11)
	return crit
end


local hitDataMOB = { 95, 94, 83, 72, 61, 50, 39, 28, 17, 6 }
local hitDataPlayer = { 95, 94, 87, 80, 73, 66, 59, 52, 45, 38 }
local lastTargetLevel
local lastSpellHit

function DrDamage:SpellHit()

	local targetLevel = UnitLevel( "target" )
	local playerLevel = UnitLevel( "player" )
	local hitPerc = 99
	
	if DrDamage.db.profile.HitTarget then
		
		local levelDiff = 0

		if playerLevel >= 60 and targetLevel == -1 then
			levelDiff = 3
		elseif targetLevel == -1 then
			levelDiff = 10
		elseif targetLevel > playerLevel then
			levelDiff = targetLevel - playerLevel
		elseif targetLevel == playerLevel then
			hitPerc = 96
		end
		
		if levelDiff > 0 and levelDiff <= 10 then
		
			if UnitIsPlayer( "target" ) then
				hitPerc = hitDataPlayer[levelDiff]
			else
				hitPerc = hitDataMOB[levelDiff]
			end
		end
		
		lastTargetLevel = targetLevel
		lastSpellHit = hitPerc
		
	elseif not DrDamage.db.profile.HitTarget then
		
		if DrDamage.db.profile.TargetLevel >= 1 then
			
			if DrDamage.db.profile.TargetPlayer then
					hitPerc = hitDataPlayer[ math.floor( DrDamage.db.profile.TargetLevel ) ]
			else
					hitPerc = hitDataMOB[ math.floor( DrDamage.db.profile.TargetLevel ) ]
			end
					
		elseif math.floor( DrDamage.db.profile.TargetLevel ) == 0 then
			hitPerc = 96
		end
		
		lastTargetLevel = DrDamage.db.profile.TargetLevel
		lastSpellHit = hitPerc		
	end
		
	return hitPerc
end

function DrDamage:GetPlusDamage( school )

	local spellDMG = GetSpellBonusDamage(1)
	
	if school == "Holy" then
  		return GetSpellBonusDamage(2)
	elseif school == "Fire" then
  		return GetSpellBonusDamage(3)
  	elseif school == "Nature" then
  		return GetSpellBonusDamage(4)
  	elseif school == "Frost" then
  		return GetSpellBonusDamage(5)
  	elseif school == "Shadow" then
  		return GetSpellBonusDamage(6)  		
   	elseif school == "Arcane" then
  		return GetSpellBonusDamage(7)
  	else
  		return spellDMG
  	end	
end


local CalculationResults = {}

function DrDamage:RawNumbers( spell, spellName )

	if not spell or not spellName then
		do return end
	end
			
	local baseSpell = (spellInfo[spellName])[0]
	
	if BS:HasReverseTranslation( spellName ) then
		spellName = BS:GetReverseTranslation( spellName )
	end		
	
	if not talents or #talents < 1 then
		DrDamage:DmgUpdateTalents()
	end
	
	local healingSpell = false
	local spellSchool 
	local spellType
	
	if type( baseSpell.School ) == "table" then
		spellSchool = baseSpell.School[1]
		spellType = baseSpell.School[2]
		
		if spellType == "Healing" then
			healingSpell = true
		end
	else
		spellSchool = baseSpell.School
	end
	
	if baseSpell.Healing then
		healingSpell = true
	end	
	
	local critPerc = DrDamage:SpellCrit() 		--Calculate crit
	local hitPerc = lastSpellHit
	
	--Calculate base hit chance only IF needed:
	local targetLevel = UnitLevel( "target" )
	
	if DrDamage.db.profile.HitTarget and targetLevel ~= lastTargetLevel then
		hitPerc = DrDamage:SpellHit() 		
	elseif not DrDamage.db.profile.HitTarget and DrDamage.db.profile.TargetLevel ~= lastTargetLevel then
		hitPerc = DrDamage:SpellHit()
	end
		
	hitPerc = hitPerc + GetCombatRatingBonus( 8 )   --Add Gear
	
	
	local spellDmg
	if healingSpell then
		spellDmg = GetSpellBonusHealing() 			-- +healing
	else
		spellDmg = DrDamage:GetPlusDamage( spellSchool ) 	--+all spelldamage
	end
	
	
	--Optional ItemBonusLib stuff
	local iBonusLib = false
	
	if IsAddOnLoaded( "ItemBonusLib" ) then
		iBonusLib = true
		if not IBL then
			IBL = AceLibrary("ItemBonusLib-1.0")
		end
	elseif IsAddOnLoadOnDemand( "ItemBonusLib" ) then
		local loaded = LoadAddOn( "ItemBonusLib" )
		if loaded then
			iBonusLib = true
			IBL = AceLibrary("ItemBonusLib-1.0")				
		end
	end
	
	if iBonusLib and IBL then
	
		if healingSpell then
			--local charMP5 = IBL:GetBonus( "MANAREG" )
		else
			--Damage to undead if target is undead
			if UnitCreatureType( "target" ) == "Undead" then
				spellDmg = spellDmg + IBL:GetBonus( "DMGUNDEAD" )
			end
		end
	end
	
	if DrDamage.db.profile.ManualDmg then
		
		if not DrDamage.db.profile.ManualPlus then
			
			if healingSpell then
				spellDmg = DrDamage.db.profile.Healing
			else
				spellDmg = DrDamage.db.profile.SpellDamage
			end
			
			critPerc = critPerc - GetCombatRatingBonus(11) + DrDamage.db.profile.ItemCrit * 100
			hitPerc = hitPerc - GetCombatRatingBonus(8) + DrDamage.db.profile.ItemHit * 100
		else
			if healingSpell then
				spellDmg = spellDmg + DrDamage.db.profile.Healing
			else
				spellDmg = spellDmg + DrDamage.db.profile.SpellDamage
			end
			
			critPerc = critPerc + DrDamage.db.profile.ItemCrit
			hitPerc = hitPerc + DrDamage.db.profile.ItemHit
		end
		
		if not healingSpell then
		
			if ( spellSchool == "Shadow" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ShadowDamage --add shadowdamage
			elseif ( spellSchool == "Fire" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FireDamage --add firedamage
			elseif ( spellSchool == "Nature" ) then
				spellDmg = spellDmg + DrDamage.db.profile.NatureDamage --add naturedamage
			elseif ( spellSchool == "Frost" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FrostDamage --add frostdamage
			elseif ( spellSchool == "Arcane" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ArcaneDamage --add arcanedamage
			elseif ( spellSchool == "Holy" ) then
				spellDmg = spellDmg + DrDamage.db.profile.HolyDamage --add holydamage			
			end
		else
			spellDmg = spellDmg + DrDamage.db.profile.Healing
		end
	end
					
	local critM = 0.5			--Crit multiplier
	local bNukeDmg = 0			--Modifier for spells with a DOT portion
	local dmgM = 1				--Damage multiplier (total)
	local bDmgM = 1				--Base damage Modifier
	local spellDmgM				--Spell damage multiplier
	local dotSpellDmgM = 0			--Dot portion spelldamage modifier
	local igniteM = 0			--Ignites % of total
	local cooldown = 0			--Spell's cooldown
	local finalMod = 0			--Modifier to final damage +/-
	local MDRank = 0			--Special variable for MD locks (the buff returns no rank)
	local minDam = spell[1]			--Spell initial base min damage
	local maxDam = spell[2]			--Spell initial base max damage
	
	if baseSpell.Cooldown then
		cooldown = baseSpell.Cooldown
	end

	local castTime = 1.5
	
	if spell.castTime then
		castTime = spell.castTime
	elseif baseSpell.castTime then
		castTime = baseSpell.castTime
	end
	
	local eDuration = 1
	
	if spell.eDuration then
		eDuration = spell.eDuration
	elseif baseSpell.eDuration then
		eDuration = baseSpell.eDuration
	end
	
	if baseSpell.bonusFactor then
		spellDmgM = baseSpell.bonusFactor
	elseif baseSpell.eDot then 
		spellDmgM = ( eDuration / 15)
	else
		spellDmgM = ( castTime / 3.5) 
	end
				
	for i, charTalent in ipairs( talents ) do
		
		local talentName = charTalent.Name
		local talentRank = charTalent.Rank
		
		if talentName then
						
			if talentEffects[talentName] then
				
				local effectTalent = talentEffects[talentName]
				
				function DoCalc( ModType, talentSpell, talentValue, add )
				
					if talentSpell == "All" or talentSpell == spellSchool or talentSpell == spellType or talentSpell == spellName then
					
						if ModType == "SpellDamage" then
						
							if not add then
								spellDmgM = spellDmgM * ( 1 + talentValue )
							else
								spellDmgM = spellDmgM + talentValue
							end
						elseif ModType == "CastTime" then
							
							if baseSpell.CastMod then
								castTime = castTime - baseSpell.CastMod * talentValue
							else
								castTime = castTime - talentValue
							end
								
							if castTime < 1.5 then
								castTime = 1.5
							end
						elseif ModType == "BaseNuke" then
							bNukeDmg = bNukeDmg + talentValue
						elseif ModType == "BaseDamage" then
							bDmgM = bDmgM + talentValue
						elseif ModType == "DotSpellDamage" then
							dotSpellDmgM = dotSpellDmgM + talentValue
						elseif ModType == "Crit" then
							critPerc = critPerc + talentValue
						elseif ModType == "Hit" then
							hitPerc = hitPerc + talentValue
						elseif ModType == "CritMultiplier" then
							critM = critM + talentValue
						elseif ModType == "Ignite" then
							igniteM = talentValue
						elseif ModType == "SpellDuration" then
							eDuration = eDuration + talentValue
						elseif ModType == "Cooldown" then
							cooldown = cooldown - talentValue
						elseif ModType == "MD" then
							MDRank = talentValue
						elseif not ModType then
						
							if not add then
								dmgM = dmgM * ( 1 + talentValue )
							else
								dmgM = dmgM + talentValue	
							end
						end
					end				
				
				end
						
				local modType = effectTalent.ModType
				local add = effectTalent.Add
				local talentValue = talentRank * effectTalent.Effect
				
				for j, talentSpell in ipairs( effectTalent.Spells ) do
					DoCalc( modType, talentSpell, talentValue, add  )
				end
									
				if effectTalent.Spells2 then
				
					modType = effectTalent.ModType2
					add = effectTalent.Add2
					
					if effectTalent.Effect2 then
						talentValue = talentRank * effectTalent.Effect2
					end
					
					for j, talentSpell in ipairs( effectTalent.Spells2 ) do
						DoCalc( modType, talentSpell, talentValue, add )
					end
				end	
			end
		end		
	end
	
	--Start calculating Buffs/Debuffs that affects damage not provided by blizzard API.
	
	local incinerateBonus = false 		--To make sure Immolate bonus to Incinerate is only added once
	local iceLanceBonus = false		--To make sure Ice Lance get's bonus only once
	
	if not healingSpell and DamageBuffs and DamageDebuffs then
	
		for buffName, apps, _, texture, rank in SEA:DebuffIter("target") do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end
	
			if DamageDebuffs[buffName] and not baseSpell.NoDebuffs then
			
				local buffData = DamageDebuffs[buffName]
			
				if buffData.School == spellSchool or buffData.School2 == spellSchool or not buffData.School then
				
					local modType = buffData.ModType
					
					if buffData.Ranks then
						local debuffRank
				
						if rank and string.len( rank ) >= 6 then
							debuffRank = tonumber( findpattern( rank, "%d+" ) )
						end
							
						if not debuffRank then
							debuffRank = buffData.Ranks
						end
						
						if not modType then
							dmgM = dmgM * ( 1 + debuffRank * buffData.Value )
						end
					elseif buffData.Apps then
						if modType == "Crit" and apps then
							critPerc = critPerc + apps * buffData.Value
						elseif not modType and apps then
							dmgM = dmgM * ( 1 + apps * buffData.Value )
						end
					elseif buffData.Texture and texture and string.find( texture, buffData.Texture ) then
						dmgM = dmgM * ( 1 + buffData.Value )
					elseif buffData.Apps2 then
						dmgM = dmgM * ( 1 + apps * buffData.Value2 )
					elseif modType == "Special" then
						if spellName == "Incinerate" and buffName == "Immolate" and not incinerateBonus then
							minDam = minDam + spell.incinerateMin
							maxDam = maxDam + spell.incinerateMax
							incinerateBonus = true
						elseif spellName == "Ice Lance" and ( buffName == "Frost Nova" or buffName == "Frostbite" ) and not iceLanceBonus then
							dmgM = dmgM * 3
							iceLanceBonus = true
						end
					elseif not modType then
						dmgM = dmgM * ( 1 + buffData.Value )
					end
				end
			end
		end
		
		for buffName, index, apps, _, rank in SEA:BuffIter("player") do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end		
		
			if DamageBuffs[buffName] and not baseSpell.NoBuffs then
			
				local buffData = DamageBuffs[buffName]
			
				if buffData.School == spellSchool or buffData.School2 == spellSchool or not buffData.School then
				
					local modType = buffData.ModType
					
					if buffData.Ranks then
					
						local buffRank
				
						if rank and string.len( rank ) >= 6 then
							buffRank = tonumber( findpattern( rank, "%d+" ) )
						end				
			
						if not buffRank then
							buffRank = buffData.Ranks
						end
			
						if not modType then
							dmgM = dmgM * ( 1 + buffRank * buffData.Value )
						end
					elseif buffData.Apps then
						if not modType and apps then
							dmgM = dmgM * ( 1 + apps * buffData.Value )
						end
					elseif modType == "BaseDamage" then
						if buffData.Spell == spellName then
							bDmgM = bDmgM + buffData.Value
						end
					elseif modType == "SpellDamage" then
						spellDmgM = spellDmgM + buffData.Value
					elseif modType == "Crit" then
						critPerc = critPerc + buffData.Value
						
						if critPerc > 100 then
							critPerc = 100
						end
					elseif modType == "Special" then
						
						if buffName == "Burning Adrenaline" then
							castTime = 1.5
							dmgM = dmgM * 2
						elseif buffName == "Master Demonologist" then
							
							if UnitCreatureFamily( "pet" ) then
								local petFamily = UnitCreatureFamily("pet")
								
								if petFamily == "Succubus" and MDRank > 0 then
									dmgM = dmgM * ( 1 + 0.02 * MDRank )				
								elseif petFamily == "Felguard" then
									dmgM = dmgM * ( 1 + 0.01 * MDRank )
								end
							end
						end						
					elseif not modType then
						dmgM = dmgM * ( 1 + buffData.Value )
					end
				end
			end		
		end
		
	elseif healingSpell and HealingTargetBuffs and HealingTargetDebuffs then
	
		for buffName, index, apps, _, rank in SEA:BuffIter("target") do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end		
	
			if HealingTargetBuffs[buffName] then
			
				local buffData = HealingTargetBuffs[buffName]
				
				if buffData.Spell == spellName or buffData.Spell == "All" then
				
					if buffData.Apps then
						if not buffData.ModType and apps then
							dmgM = dmgM * ( 1 + apps * buffData.Value )
						end
					elseif buffData.Ranks then
						local buffRank
				
						if rank and string.len( rank ) >= 6 then
							buffRank = tonumber( findpattern( rank, "%d+" ) )
						end
						
						if not buffRank then
							buffRank = buffData.Ranks
						end
						
						if buffData.ModType == "Amplify Magic" then
							spellDmg = spellDmg + buffData.Value[buffRank]
						elseif not buffData.ModType then
							dmgM = dmgM * ( 1 +  buffRank * buffData.Value )
						end					
					elseif not buffData.ModType then
						dmgM = dmgM * ( 1 + buffData.Value )
					end					
				end
			end
		end
		
		for buffName, apps, _, _, rank in SEA:DebuffIter("target") do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end		
			
			if HealingTargetDebuffs[buffName] and not baseSpell.NoDebuffs then
			
				local buffData = HealingTargetDebuffs[buffName]
				
				if buffData.Spell == spellName or buffData.Spell == "All" then
				
					if buffData.Apps then
					
						if buffData.ModType == "Wound Poison" then
							local buffRank
				
							if rank and string.len( rank ) >= 6 then
								buffRank = tonumber( findpattern( rank, "%d+" ) )
							end
							
							if not buffRank then
								buffRank = buffData.Ranks
							end
							
							if apps then
								finalMod = -1 * apps * buffData.Value[buffRank]
							end
							
						elseif not buffData.ModType and apps then
							dmgM = dmgM * ( 1 - apps * buffData.Value )
						end
					elseif not buffData.ModType then
						dmgM = dmgM * ( 1 - buffData.Value )
					end
				end
			end
		end
	end
	
	--Modified factors?
	
	--Split between bonus to dot and nuke
	if baseSpell.hybridFactor then
		spellDmgM = spellDmgM * baseSpell.hybridFactor
	end
	
	--Factor that applies to all ranks of the spell
	if baseSpell.sFactor then
		spellDmgM = spellDmgM * baseSpell.sFactor
	end
	
	--Factor that applies to current rank of the spell
	if spell.sFactor then
		spellDmgM = spellDmgM * spell.sFactor
	end
	
	
	--Factor for downranking
	local lowRankFactor = 1
	
	if spell.maxLevel and spell.maxLevel < 68 then
		
		local playerLevel = UnitLevel("player")
		
			local downRankMod = 0
			
			if spell.Downrank then
				downRankMod = spell.Downrank
			elseif baseSpell.Downrank then
				downRankMod = baseSpell.Downrank
			end
	
			if healingSpell then
				lowRankFactor = ( spell.maxLevel + downRankMod + 6 ) / playerLevel
			else
				lowRankFactor = ( spell.maxLevel + downRankMod + 4 ) / playerLevel
			end
		
			if lowRankFactor < 1 then
				spellDmgM = spellDmgM * lowRankFactor
			else
				lowRankFactor = 1
			end
	end
	
	--Factor for spells under level 20
	local lowLevelFactor = 1
	
	if spell.spellLevel and spell.spellLevel < 20 then
	
		lowLevelFactor = 1 - ((20 - spell.spellLevel) * 0.0375)
		
		if lowLevelFactor < 1 and lowLevelFactor > 0 then
			spellDmgM = spellDmgM * lowLevelFactor
		else
			lowLevelFactor = 1
		end
	end
	
	--Adding to spells base damage after levelups:
		
	if baseSpell.BaseIncrease and spell.spellLevel then
	
		local playerLevel = UnitLevel("player")
	
		if playerLevel > spell.spellLevel then
				
			if spell.maxLevel and playerLevel >= spell.maxLevel then
				minDam = minDam + spell[3]
				maxDam = maxDam + spell[4]
				
			elseif baseSpell.NextRank and playerLevel >= ( spell.spellLevel + baseSpell.NextRank - 1 ) then
				minDam = minDam + spell[3]
				maxDam = maxDam + spell[4]			
			else
				local diff = 1
				
				if baseSpell.NextRank then
					diff = baseSpell.NextRank - 1
				elseif spell.maxLevel then
					diff = spell.maxLevel - spell.spellLevel
				end
				
				minDam = minDam + ( playerLevel - spell.spellLevel ) * ( spell[3] / diff )
				maxDam = maxDam + ( playerLevel - spell.spellLevel ) * ( spell[4] / diff )
			end
		end
		
		minDam = DrDamage:Round( minDam )
		maxDam = DrDamage:Round( maxDam )
	end
	
	function DmgCalc( next, hitCalc )
	
		--Initialize variables that are calculated
		local calcAvgDmgCrit = 0		--Calculated final damage with crit
		local calcAvgDmg = 0			--Calculated final damage without crit
		local calcMinDmg = 0			--Calculated dmg range: MIN
		local calcMaxDmg = 0			--Calculated dmg range: MAX
		local calcMinCrit = 0
		local calcMaxCrit = 0		
		local calcDotDmg = 0			--Calculated DOT portion (if applicable)
		local igniteDotDmg = 0			--Calculated Ignite DOT Portion
		local avgIgnite	= 0			--Calculated Average Ignite
		local dispSpellDmgM = spellDmgM		--Calculated spelldmg M
		local perHit = 0
		local perTick = 0
	
		
		if baseSpell.sHits then
			calcMinDmg = dmgM * ( bDmgM * minDam * baseSpell.sHits + (  spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * maxDam * baseSpell.sHits + (  spellDmg * spellDmgM ) )
		elseif spell.hybridDotDmg then
			calcMinDmg = ( dmgM + bNukeDmg ) * ( bDmgM * minDam + ( spellDmg * spellDmgM ) )
			calcMaxDmg = ( dmgM + bNukeDmg ) * ( bDmgM * maxDam + ( spellDmg * spellDmgM ) )
		else
			calcMinDmg = dmgM * ( bDmgM * minDam + ( spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * maxDam + ( spellDmg * spellDmgM ) )
		end
				
		--EXAMPLE: Wound Poison
		if finalMod ~= 0 then
		
			calcMinDmg = calcMinDmg + finalMod
			calcMaxDmg = calcMaxDmg + finalMod
			
			if calcMinDmg < 0 then
				calcMinDmg = 0
			end
			
			if calcMaxDmg < 0 then
				calcMaxDmg = 0
			end	
		end
		
		calcAvgDmg = (calcMinDmg + calcMaxDmg) / 2
				
		--This is for Imp SW:P
		if baseSpell.eDot and baseSpell.eDuration and eDuration > baseSpell.eDuration then
			local bonusDotDamage = ( calcAvgDmg / baseSpell.eDuration ) * ( eDuration - baseSpell.eDuration )
			calcAvgDmg = calcAvgDmg + bonusDotDamage
		end
		
		
		--Crit calculation:
		local critBonus = 0
		
		if critPerc > 100 then
			critPerc = 100
		end
		
		if baseSpell.canCrit then
			calcMinCrit = calcMinDmg + calcMinDmg * critM
			calcMaxCrit = calcMaxDmg + calcMaxDmg * critM
			critBonus = ( critPerc / 100 ) * calcAvgDmg * critM
			calcAvgDmgCrit = calcAvgDmg + critBonus
		else
			calcAvgDmgCrit = calcAvgDmg
			critPerc = 0;
		end

		--Hit calculation:
		if hitPerc > 99 then
			hitPerc = 99
		end
		
		if not healingSpell and not baseSpell.Unresistable then
			if DrDamage.db.profile.HitCalc or hitCalc then
				calcAvgDmgCrit = ( calcAvgDmgCrit - critBonus ) * ( hitPerc / 100 ) + critBonus
			end
		end		
					
		if spell.hybridDotDmg then
	
			local hybridFactor = 0
			local sFactor = 1
			
				if baseSpell.hybridFactor then
					hybridFactor = baseSpell.hybridFactor
				end
				
				if baseSpell.sFactor then
					sFactor = baseSpell.sFactor * sFactor
				end
				
				if spell.sFactor then
					sFactor = spell.sFactor * sFactor
				end
				
				if baseSpell.hybridDotFactor then
					sFactor = baseSpell.hybridDotFactor * sFactor
				end
		
			calcDotDmg = dmgM * ( spell.hybridDotDmg + ( eDuration / 15 + dotSpellDmgM ) * ( 1 - hybridFactor ) * sFactor * lowRankFactor * lowLevelFactor * spellDmg )
			dispSpellDmgM = dispSpellDmgM +  ( eDuration / 15 + dotSpellDmgM ) * ( 1 - hybridFactor ) * sFactor * lowRankFactor * lowLevelFactor
		end
		
		if baseSpell.Ignite then
			igniteDotDmg = igniteM * calcAvgDmgCrit
			avgIgnite = ( critPerc / 100 ) * igniteDotDmg
		end
		
		if spell.extraDotDmg then
		
			local xFactor = 0
			
			if baseSpell.extraDotF then
				xFactor = baseSpell.extraDotF
			end
			
			calcDotDmg = calcDotDmg + spell.extraDotDmg + xFactor * lowRankFactor * lowLevelFactor * spellDmg --TODO: Verify factor
			dispSpellDmgM = dispSpellDmgM + xFactor * lowRankFactor * lowLevelFactor
		end
		
		local chainEffect = 0
		
		if baseSpell.chainFactor then
			chainEffect = calcAvgDmgCrit * ( baseSpell.chainFactor + math.pow( baseSpell.chainFactor, 2 ) + math.pow( baseSpell.chainFactor, 3 ) )
		end
					
		--Begin DPS calculations
		
		local calcDPS = 0
		local calcDPT = 0
		local calcDPSCooldown = 0
	
		if baseSpell.eDot then
			calcDPS = calcAvgDmgCrit / eDuration
		else
			if baseSpell.Ignite then
			
				calcDPS = calcAvgDmgCrit / castTime + avgIgnite / 4
				
				if spell.hybridDotDmg or spell.extraDotDmg then
					calcDPS = calcDPS + calcDotDmg / eDuration
				end
				
				calcAvgDmgCrit = calcAvgDmgCrit + avgIgnite + calcDotDmg
				
			elseif spell.hybridDotDmg or spell.extraDotDmg then
				calcDPS = calcAvgDmgCrit / castTime + calcDotDmg / eDuration
				calcAvgDmgCrit = calcAvgDmgCrit + calcDotDmg
			else
				calcDPS = calcAvgDmgCrit / castTime
			end
		end
		
		calcDPT = calcAvgDmgCrit / castTime
		
		if baseSpell.Cooldown then
			
			if baseSpell.eDot then
				calcDPSCooldown = calcAvgDmgCrit / cooldown
			else
				calcDPSCooldown = calcAvgDmgCrit / cooldown
			end
		end
		
		local AvgTotal = 		calcAvgDmgCrit		
		local DPS = 			DrDamage:Round( calcDPS, 1 )
		local DPSC = 			DrDamage:Round( calcDPT, 1 )
		local Avg =			DrDamage:Round( calcAvgDmg )
			
		if not next then
			
			CalculationResults.MinDmg = 	DrDamage:Round( calcMinDmg )
			CalculationResults.MaxDmg = 	DrDamage:Round( calcMaxDmg )
			CalculationResults.MinCrit =	DrDamage:Round( calcMinCrit )
			CalculationResults.MaxCrit = 	DrDamage:Round( calcMaxCrit )
			CalculationResults.AvgCrit = 	DrDamage:Round( ( calcMinCrit + calcMaxCrit ) / 2 )
			CalculationResults.CritRate = 	DrDamage:Round( critPerc, 2 )
			CalculationResults.HitRate = 	DrDamage:Round( hitPerc, 2 )
			CalculationResults.DotDmg = 	DrDamage:Round( calcDotDmg )
			CalculationResults.SpellDmg = 	DrDamage:Round( spellDmg )
			CalculationResults.SpellDmgM = 	DrDamage:Round( dispSpellDmgM, 3 )
			CalculationResults.DmgM = 	DrDamage:Round( dmgM, 3 )
			CalculationResults.IgniteDmg = 	DrDamage:Round( igniteDotDmg )
			CalculationResults.DPSCD = 	DrDamage:Round( calcDPSCooldown, 1 )
			CalculationResults.ChainDmg = 	DrDamage:Round( chainEffect )
			
			local hits
			local ticks
		
			if spell.sHits then
				hits = spell.sHits
				perHit = calcAvgDmg / hits
			elseif baseSpell.sHits then
				hits = baseSpell.sHits
				perHit = calcAvgDmg / hits	
			end
		
			if spell.sTicks then
				
				if castTime > eDuration then
					ticks = castTime / spell.sTicks
				else
					ticks = eDuration / spell.sTicks
				end
				
				if spell.hybridDotDmg or spell.extraDotDmg then
					perTick = calcDotDmg / ticks
				else
					perTick = calcAvgDmg / ticks
				end
			elseif baseSpell.sTicks then
			
				if castTime > eDuration then
					ticks = castTime / baseSpell.sTicks
				else
					ticks = eDuration / baseSpell.sTicks
				end			
				
				if spell.hybridDotDmg or spell.extraDotDmg then
					perTick = calcDotDmg / ticks
				else
					perTick = calcAvgDmg / ticks
				end
			end		
		
			if perHit > 0 then
				CalculationResults.PerHit = { DrDamage:Round( perHit, 1 ), hits }
			end
		
			if perTick > 0 then
				CalculationResults.PerTick = { DrDamage:Round( perTick, 1 ), ticks }
			end			
				
		
		end
		
		return AvgTotal, DPS, DPSC, Avg
	end
	
	local returnAvgTotal, returnDPS, returnDPSC, returnAvg = DmgCalc()
	
	local returnNextCrit = 0
	local returnNextTenDmg = 0
	local returnNextHit = 0
	
	if DrDamage.db.profile.Next then
	
		if baseSpell.canCrit then
			critPerc = critPerc + 1
			returnNextCrit = DrDamage:Round( DmgCalc( true ) - returnAvgTotal, 1 )
			critPerc = critPerc - 1	
		end
	
		spellDmg = spellDmg + 10
		returnNextTenDmg = DrDamage:Round( DmgCalc( true ) - returnAvgTotal, 1 )
		spellDmg = spellDmg - 10
	
		if not healingSpell then
			local temp = DmgCalc( true, true )
			hitPerc = hitPerc + 1
			returnNextHit = DrDamage:Round( DmgCalc( true, true ) - temp, 1 )
			hitPerc = hitPerc - 1
		end
	end
	
	returnAvgTotal = DrDamage:Round( returnAvgTotal )
	
	CalculationResults.NextCrit = returnNextCrit
	CalculationResults.NextTenDmg = returnNextTenDmg
	CalculationResults.NextHit = returnNextHit
	
	return healingSpell, returnAvgTotal, returnDPS, returnDPSC, returnAvg, CalculationResults
end

function DrDamage:Round(x, y)

	local s
	
	if y then 
		x = x * math.pow(10, y) 
	end
	
	if (x > 0) then
		s = math.floor(x+0.5);
  	else
    		s = math.ceil(x-0.5);
  	end
  	
  	if y then 
  		s = s / math.pow(10, y) 
  	end
  	
  	return s
end

